#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import unittest

import mesh.log as log
import mesh.tool as tool
from mesh.kinds import Service


class TestTools(unittest.TestCase):

    def test_abstract_class(self):
        service = Service()
        service.address = '1'
        log.info(service.address)

    def test_new_trace_id(self):
        log.info(tool.new_trace_id())
        log.info(tool.get_ip())

    def test_required(self):
        log.info(f"{tool.required(True)}")
        log.info(f"{tool.required(True, False)}")
        log.info(f"{tool.required(0, 0)}")
        log.info(f"{tool.required([])}")
        log.info(f"{tool.required([], [])}")
        log.info("False begin")
        log.info(f"{tool.required()}")
        log.info(f"{tool.required(1, None)}")
        log.info(tool.get_ip())
        log.info(tool.get_ip())


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import spi
from mesh.prsim import RuntimeHook


@spi("mesh")
class MeshRuntimeHook(RuntimeHook):

    def start(self):
        """"""
        pass

    def stop(self):
        """"""
        pass

    def refresh(self):
        """"""
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from itertools import repeat
from typing import List

import mesh.log as log

MESH_DOMAIN = "trustbe.cn"


class URN:
    """
    Like:

    create.tenant.omega . 0001 . json . http2 . lx000001 . trustbe.net
    -------------------   ----   ----   -----   --------   -----------
    name            flag   codec  proto     node       domain

    """

    @staticmethod
    def parse(urn: str) -> "URN":
        instance = URN()
        if urn is None or urn.isspace():
            log.warn(f'Unresolved urn {urn}')
            return instance
        names = URN.as_array(urn)
        if len(names) < 5:
            log.warn(f'Unresolved urn {urn}')
            return instance
        names.reverse()
        instance.name = "{}.{}".format(names[1], names[0])
        instance.node_id = names[2]
        instance.flag = URNFlag.parse(names[3])
        instance.name = ".".join(names[4:])
        return instance

    def __init__(self):
        self.domain = MESH_DOMAIN
        self.node_id = ""
        self.flag = URNFlag()
        self.name = ""

    def __str__(self):
        return self.string()

    def match_node(self, node_id: str) -> bool:
        return self.node_id.lower() == node_id.lower()

    def string(self):
        urn = self.as_array(self.name)
        urn.reverse()
        urn.append(self.flag.string())
        urn.append(self.node_id.lower())
        if self.domain.isspace():
            urn.append(MESH_DOMAIN)
        else:
            urn.append(self.domain.lower())
        return ".".join(urn)

    @staticmethod
    def as_array(urn: str) -> List[str]:
        pairs = urn.split(".")
        names = []
        buffer = []
        for pair in pairs:
            if pair and pair.startswith("${"):
                buffer.append(pair)
                buffer.append('.')
                continue
            if pair and pair.endswith("}"):
                buffer.append(pair)
                names.append(''.join(buffer))
                buffer.clear()
                continue
            names.append(pair)
        return names


class URNFlag:

    @staticmethod
    def parse(value: str):
        flags = URNFlag()
        if value is None or value.isspace():
            log.error(f'Unresolved Flag {value}')
            return flags
        flags.v = value[0:2]
        flags.proto = value[2:4]
        flags.codec = value[4:6]
        flags.version = URNFlag.reduce(value[6:12], 2)
        flags.zone = value[12:14]
        flags.cluster = value[14:16]
        flags.cell = value[16:18]
        flags.group = value[18:20]
        flags.address = URNFlag.reduce(value[20:32], 3)
        flags.port = URNFlag.reduce(value[32:37], 5)
        return flags

    @staticmethod
    def reduce(value: str, length: int):
        buffer = []
        for index, code in enumerate(value):
            if (index + length) % length != 0:
                continue
            has_none_zero = True
            for offset in range(index, index + length):
                if value[offset] != '0':
                    has_none_zero = False
                if not has_none_zero:
                    buffer.append(value[offset])
            if has_none_zero:
                buffer.append('0')
            buffer.append('.')

        return ''.join(buffer[0:len(buffer) - 1])

    def __init__(self):
        self.v = '00'  # #                 2      00
        self.proto = '00'  # #             2      00
        self.codec = '00'  # #             2      00
        self.version = '000000'  # #       6      000000
        self.zone = '00'  # #              2      00
        self.cluster = '00'  # #           2      00
        self.cell = '00'  # #              2      00
        self.group = '00'  # #             2      00
        self.address = '000000000000'  # # 12     000000000000
        self.port = '00080'  # #           5      00080   Max port 65535

    @staticmethod
    def padding(v: str, length: int) -> str:
        value = v.replace(".", "")
        if len(value) == length:
            return value
        if len(value) < length:
            return f'{"".join(repeat("0", length - len(v)))}{value}'
        return v[0:length]

    def padding_chain(self, v: str, length: int, size: int) -> str:
        def chain(values: [str]) -> str:
            chars = []
            for frag in values:
                chars.append(self.padding(frag, length))
            return "".join(chars)

        frags = v.split(".")
        minimal = len(frags)
        if len(frags) == size:
            return chain(frags)
        if len(frags) < size:
            for _ in range(0, size - minimal):
                frags.append("")
            return chain(frags)
        return chain(frags[0:size])

    def string(self):
        return "{}{}{}{}{}{}{}{}{}{}".format(
            self.padding(self.v, 2),
            self.padding(self.proto, 2),
            self.padding(self.codec, 2),
            self.padding_chain(self.version, 2, 3),
            self.padding(self.zone, 2),
            self.padding(self.cluster, 2),
            self.padding(self.cell, 2),
            self.padding(self.group, 2),
            self.padding_chain(self.address, 3, 4),
            self.padding(self.port, 5)
        )

    def reset(self, proto: str, codec: str, version: str, zone: str, cluster: str, cell: str, group: str, address: str,
              port: str) -> "URNFlag":
        self.proto = proto
        self.codec = codec
        self.version = version
        self.zone = zone
        self.cluster = cluster
        self.cell = cell
        self.group = group
        self.address = address
        self.port = port
        return self



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro.ark import ark, T, A, B, C
from mesh.macro.binding import binding, Binding
from mesh.macro.cause import Cause
from mesh.macro.codec import serializable
from mesh.macro.compatible import Compatible
from mesh.macro.index import index, Index
from mesh.macro.inspect import Returns, Parameters, Inspector
from mesh.macro.loader import ServiceLoader
from mesh.macro.mpi import mpi, MPI, MethodProxy
from mesh.macro.mps import mps, MPS
from mesh.macro.proxy import Proxy, InvocationHandler
from mesh.macro.spi import spi, SPI
from mesh.macro.types import Types

__all__ = ("mpi",
           "mps",
           "binding",
           "index",
           "spi",
           "serializable",
           "ark",
           "T",
           "A",
           "B",
           "C",
           "Index",
           "SPI",
           "Proxy",
           "InvocationHandler",
           "MPI",
           "MPS",
           "Binding",
           "ServiceLoader",
           "Types",
           "Cause",
           "Returns",
           "Parameters",
           "Inspector",
           "MethodProxy",
           "Compatible",
           )


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Dict, Any, Callable

import mesh.log as log
import mesh.tool as tool
from mesh.kinds import Principal
from mesh.macro import spi
from mesh.mpc import ServiceProxy, Mesh
from mesh.prsim import Transport, Session, Metadata


@spi("mesh")
class MeshTransport(Transport):
    """
    Mesh transport is stateless transport depends on mesh server.
    """

    def __init__(self):
        self.sessions: Dict[str, Session] = dict()

    def open(self, session_id: str, metadata: Dict[str, str]) -> Session:
        skey = self.session_key(session_id, metadata)
        session = self.sessions.get(skey, None)
        if not session:
            attachments = Mesh.context().get_attachments()
            addr = Mesh.context().get_attribute(Mesh.REMOTE)
            principal = Mesh.context().get_principals().peek()
            session = MeshSession(session_id, metadata, attachments, addr, principal, lambda: self.finalize(skey))
            self.sessions[skey] = session
        return session

    def close(self, timeout: int):
        for k, v in self.sessions.items():
            try:
                v.release(timeout, "")
            except BaseException as e:
                log.error("", e)
        self.sessions.clear()

    def roundtrip(self, payload: bytes, metadata: Dict[str, str]) -> bytes:
        """ bid """
        pass

    def finalize(self, session_key: str):
        self.sessions.__delitem__(session_key)

    @staticmethod
    def session_key(session_id: str, metadata: Dict[str, str]) -> str:
        target_node_id = Metadata.MESH_TARGET_NODE_ID.get(metadata)
        target_inst_id = Metadata.MESH_TARGET_INST_ID.get(metadata)
        return f"{session_id}.{target_node_id}.{target_inst_id}"


@spi("mesh")
class MeshSession(Session):

    def __init__(self, session_id: str, metadata: Dict[str, str], attachments: Dict[str, str], address: str,
                 principal: Principal, finalizer: Callable):
        self.session_id = session_id
        self.metadata = metadata
        self.attachments = attachments
        self.address = address
        self.principal = principal
        self.session = ServiceProxy.default_proxy(Session)
        self.finalizer = finalizer

    def with_context(self, fn: Any, remote: bool, timeout: int) -> Any:
        return Mesh.context_safe(lambda: self.context_safe(fn, remote, timeout))

    def context_safe(self, fn: Any, remote: bool, timeout: int) -> Any:
        if self.attachments:
            for k, v in self.attachments.items():
                Mesh.context().get_attachments()[k] = v

        if self.metadata:
            for k, v in self.metadata.items():
                Mesh.context().get_attachments()[k] = v

        if tool.required(self.address):
            Mesh.context().set_attribute(Mesh.REMOTE, self.address)

        if timeout > 0:
            Mesh.context().set_attribute(Mesh.TIMEOUT, timeout)

        Metadata.MESH_SESSION_ID.set(Mesh.context().get_attachments(), self.session_id)

        if not remote:
            return fn()

        target_node_id = Metadata.MESH_TARGET_NODE_ID.get(Mesh.context().get_attachments())
        target_inst_id = Metadata.MESH_TARGET_INST_ID.get(Mesh.context().get_attachments())
        if '' == target_node_id and '' == target_inst_id:
            return fn()
        try:
            principal = Principal()
            principal.node_id = target_node_id
            principal.inst_id = target_inst_id
            Mesh.context().get_principals().append(principal)
            return fn()
        finally:
            Mesh.context().get_principals().pop()

    def peek(self, topic: str = "") -> bytes:
        return self.with_context(lambda: self.session.peek(topic), False, 0)

    def pop(self, timeout: int, topic: str = "") -> bytes:
        return self.with_context(lambda: self.session.pop(timeout, topic), False, timeout)

    def push(self, payload: bytes, metadata: Dict[str, str], topic: str = ""):
        return self.with_context(lambda: self.session.push(payload, metadata, topic), True, 0)

    def release(self, timeout: int, topic: str = ""):
        self.finalizer()
        return self.with_context(lambda: self.session.release(timeout, topic), False, timeout)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import signal
import threading

import mesh.asm as asm
from mesh.boost import *
from mesh.cause import *
from mesh.codec import *
from mesh.context import *
from mesh.macro import *
from mesh.mpc import *
from mesh.prsim import *

__all__ = (
    "mpi",
    "mps",
    "index",
    "spi",
    "binding",
    "Codeable",
    "Cause",
    "Inspector",
    "Types",
    "ServiceLoader",
    ##
    "URN",
    "URNFlag",
    "Consumer",
    "Filter",
    "Invocation",
    "Provider",
    "Schema",
    "ServiceProxy",
    "MeshKey",
    "Mesh",
    # cause
    "MeshException",
    "CompatibleException",
    "NotFoundException",
    "ValidationException",
    "Codec",
    # prsim
    "Builtin",
    "Cache",
    "Cluster",
    "Commercialize",
    "RunMode",
    "Key",
    "Metadata",
    "Queue",
    "Context",
    "Cryptor",
    "DataHouse",
    "Dispatcher",
    "Endpoint",
    "EndpointSticker",
    "Evaluator",
    "Graphics",
    "Hodor",
    "IOStream",
    "KV",
    "Licenser",
    "Locker",
    "Network",
    "Publisher",
    "Registry",
    "Routable",
    "RuntimeHook",
    "Scheduler",
    "Sequence",
    "Subscriber",
    "Tokenizer",
    "Transport",
    "Session",
    #
    "Runtime",
    "MethodProxy",
    "Metadata",
)

__mesh_runtime__ = Mooter()

from mesh.prsim.graphics import Graphics


def init():
    asm.init()


def start():
    __mesh_runtime__.start()


def refresh():
    __mesh_runtime__.refresh()


def stop():
    __mesh_runtime__.stop()


def wait():
    """
    Use signal handler to throw exception which can be caught to allow graceful exit.
    :return:
    """
    waiter = threading.Event()

    def signal_handler(signum, frame):
        try:
            stop()
        finally:
            waiter.set()

    signal.signal(signal.SIGTERM, signal_handler)
    waiter.wait()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.codec.codec import Codec, T
from mesh.macro.spi import spi

Protobuf = "protobuf"


@spi(Protobuf)
class ProtobufCodec(Codec):

    def encode(self, value: T) -> bytes:
        pass

    def decode(self, value: bytes) -> T:
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import sys

try:
    if sys.version_info < (3, 8):
        import importlib_metadata as metadata
    else:
        from importlib import metadata
except (ImportError, OSError):
    metadata = None


def _get_version():
    default_version = '0.0.1'
    try:
        version = metadata.version('mesh')
    except (AttributeError, ImportError, OSError):
        version = default_version
    return version


__version__ = _get_version()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from concurrent.futures import Future

from mesh.kinds import Reference
from mesh.macro import spi
from mesh.mpc import Consumer, Execution


@spi("http")
class HTTPConsumer(Consumer):

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def start(self):
        pass

    def consume(self, address: str, urn: str, execution: Execution[Reference], inbound: bytes) -> Future:
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import inspect
from concurrent.futures import Future, CancelledError
from typing import Any, Type, Dict

import mesh.log as log
import mesh.tool as tool
from mesh.cause import MeshException, NotFoundException, CompatibleException, MeshCode
from mesh.codec import Codec
from mesh.context import Mesh, MeshKey
from mesh.kinds import Reference
from mesh.macro import mpi, InvocationHandler, T, ServiceLoader, Parameters, Inspector
from mesh.mpc.consumer import Consumer
from mesh.mpc.eden import Eden
from mesh.mpc.filter import Filter, CONSUMER
from mesh.mpc.inspector import MethodInspector
from mesh.mpc.invoker import Invoker, Invocation, ServiceInvocation, Execution
from mesh.mpc.urn import URN
from mesh.prsim import Key

# Mesh invocation attributes.
INVOCATION: Key[Invocation] = MeshKey("mesh.invocation", str)


class ReferenceInvokeHandler(Invoker, InvocationHandler):
    """
    ReferenceInvokeHandler is mesh remote call.
    """

    def __init__(self, macro: mpi, kind: Type[T]):
        inspectors: Dict[Any, Inspector] = {}
        for interface in kind.__mro__:
            if not inspect.isabstract(interface):
                continue
            for kind, methods in tool.get_declared_methods(interface).items():
                for method in methods:
                    inspectors[method] = MethodInspector(macro, interface, method)
        self.macro = macro
        self.kind = kind
        self.invoker = Filter.composite(self, CONSUMER)
        self.inspectors = inspectors

    def invoke(self, proxy: Any, method: Any, *args, **kwargs):
        return Mesh.context_safe(lambda: self._invoke(proxy, method, *args, **kwargs))

    def _invoke(self, proxy: Any, method: Any, *args, **kwargs):
        inspector = self.inspectors.get(method)
        execution: Execution[Reference] = self.refer_execution(inspector)
        urn = self.rewrite_urn(execution)
        Mesh.context().rewrite_urn(urn)
        Mesh.context().set_attribute(Mesh.REMOTE, self.rewrite_address(urn))

        parameters: Parameters = execution.inflect()
        parameters.set_arguments(*args)
        parameters.set_attachments({})

        invocation = ServiceInvocation(self, inspector, parameters, execution, URN.parse(urn))

        Mesh.context().set_attribute(INVOCATION, invocation)
        try:
            return self.invoker.run(invocation)
        except BaseException as e:
            log.error(f'Invoke {Mesh.context().get_urn()} fault because of {e}')
            raise e

    def run(self, invocation: Invocation) -> Any:
        execution = self.refer_execution(invocation.get_inspector())
        consumer = ServiceLoader.load(Consumer).get_default()
        address = tool.anyone(Mesh.context().get_attribute(Mesh.REMOTE), tool.get_mesh_address().any())
        name = execution.schema().codec
        codec = ServiceLoader.load(Codec).get(name)
        buff = codec.encode(invocation.get_parameters())
        future = consumer.consume(address, Mesh.context().get_urn(), execution, buff)
        if invocation.is_futures():
            return self.deserialize(execution, codec, future)
        return self.deserialize(execution, codec, future)

    @staticmethod
    def deserialize(execution: Execution[Reference], codec: Codec, future: Future) -> Any:
        try:
            output = future.result()
            returns = codec.decode(output, execution.retype())
            if returns.get_cause():
                log.error(returns.get_cause().text)
                raise MeshException(returns.get_code(), returns.get_message())
            if MeshCode.NOT_FOUND.matches(returns.get_code()):
                raise NotFoundException(returns.get_message())
            if not MeshCode.SUCCESS.matches(returns.get_code()):
                raise MeshException(returns.get_code(), returns.get_message())
            if not execution.inspect().get_return_type():
                return None
            return returns.get_content()
        except CancelledError as e:
            log.error(f'Invoke {Mesh.context().get_urn()} fault because of {e}')
            raise e.__cause__
        except BaseException as e:
            log.error(f'Invoke {Mesh.context().get_urn()} fault because of {e}')
            raise e

    def refer_execution(self, inspector: Inspector) -> Execution[Reference]:
        eden = ServiceLoader.load(Eden).get_default()
        execution = eden.refer(self.macro, inspector.get_type(), inspector)
        if execution:
            return execution
        raise CompatibleException(f"Method {inspector.get_name()} cant be compatible")

    @staticmethod
    def rewrite_urn(execution: Execution[Reference]) -> str:
        urn = execution.schema().urn
        principal = Mesh.context().get_principals().peek()
        inst_id = '' if not principal else principal.inst_id
        node_id = '' if not principal else principal.node_id
        remote_name = Mesh.context().get_attribute(Mesh.REMOTE_NAME)
        uname = Mesh.context().get_attribute(Mesh.UNAME)
        if tool.optional(node_id) and tool.optional(inst_id) and tool.optional(remote_name) and tool.optional(uname):
            return urn
        urn = URN.parse(urn)
        if tool.required(inst_id):
            urn.node_id = principal.inst_id
        if tool.required(node_id):
            urn.node_id = principal.node_id
        if tool.required(uname):
            urn.name = uname
        if tool.required(remote_name):
            urn.name = urn.name.replace("${mesh.name}", remote_name)
        return str(urn)

    def rewrite_address(self, uns: str):
        caddr = Mesh.context().get_attribute(Mesh.REMOTE)
        if tool.required(caddr):
            return caddr
        urn = URN.parse(uns)
        if urn.name and urn.name.startswith(".mesh"):
            return tool.get_mesh_address().any()
        if tool.get_mesh_direct():
            names = tool.get_mesh_direct().split(",")
            for name in names:
                pair = name.split("=")
                if self.is_direct(urn, pair):
                    return pair[1]
        address = urn.flag.address.replace(".", "")
        if address.isnumeric() and int(address) > 0:
            return f'{urn.flag.address}:{urn.flag.port}'
        return tool.get_mesh_address().any()

    @staticmethod
    def is_direct(urn: URN, pair: []) -> bool:
        if pair.__len__() < 2 or tool.optional(pair[1]):
            return False
        if not pair[0].__contains__("@"):
            return urn.name.startswith(pair[0])
        nn = pair[0].split("@")
        if nn.__len__() < 2 or tool.optional(nn[1]):
            return False
        return urn.node_id.lower() == nn[1].lower() and urn.name == nn[0]



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


import inspect
from abc import ABC, abstractmethod
from typing import Any, Type

from mesh.macro.ark import ark, T
from mesh.macro.loader import ServiceLoader
from mesh.macro.spi import spi


@spi("mpc")
class MethodProxy(ABC):

    @abstractmethod
    def proxy(self, kind: Type[T]) -> T:
        """Proxy facade"""
        pass


class MPI:

    def __init__(self, fn: str = None, name='', version='', proto='grpc', codec='json',
                 asyncable=True, timeout=10000, retries=3, node='', inst='', zone='', cluster='', cell='', group='',
                 address='', encrypt=False):
        """
        Multi Provider Interface. Mesh provider interface.

        :param name: service name.
        :param version: service version.
        :param proto: invoke net protocol.
        :param codec: invoke serialize protocol.
        :param asyncable: invoke asyncable.
        :param timeout: invoke timeout with mills.
        :param retries: invoke retry times.
        :param node: invoke to any network node.
        :param inst: alias of node.
        :param zone: invoke to network zone.
        :param cluster: invoke to network cluster.
        :param cell: invoke to network cell.
        :param group: invoke to network group.
        :param address: invoke to direct address.
        """
        self.__proxy__ = None
        self.urn = ""
        self.namespace = ""
        self.name = name
        self.version = version
        self.proto = proto
        self.codec = codec
        self.asyncable = asyncable
        self.timeout = timeout
        self.retries = retries
        self.node = node
        self.inst = inst
        self.zone = zone
        self.cluster = cluster
        self.cell = cell
        self.group = group
        self.address = address
        self.encrypt = encrypt
        if fn:
            fn_is_str = type(fn) is str
            fn_qualname = getattr(fn, '__qualname__', '').split(".")
            fn_name = fn_qualname[0] if fn_qualname.__len__() < 2 else fn_qualname[1]
            fn_namespace = '' if fn_qualname.__len__() < 2 else fn_qualname[0]
            self.fn = None if fn_is_str else fn
            self.declared_kind = None if fn_is_str else ark.get_declared_class(self.fn)
            self.namespace = '' if fn_is_str else fn_namespace
            self.name = fn if fn_is_str else fn_name

    def __call__(self, *args, **kwargs):
        """ Interpreter is executing when called, so some metadata must get delay """
        if not getattr(self, 'fn', None):
            self.fn = args[0]
            self.declared_kind = ark.get_declared_class(self.fn)
        if self.declared_kind or inspect.isabstract(self.declared_kind):
            signature = inspect.signature(self.fn)
            self.__proxy__ = ServiceLoader.load(MethodProxy).get_default().proxy(signature.return_annotation)

        ark.register(mpi, self.name, self.fn, self)
        self.fn.__mpi__ = self
        return self.fn

    def __get__(self, ref, kind):
        return self.proxy_if_absent()

    def proxy_if_absent(self):
        if not self.declared_kind or not inspect.isabstract(self.declared_kind):
            self.declared_kind = ark.get_declared_class(self.fn)
            signature = inspect.signature(self.fn)
            self.__proxy__ = ServiceLoader.load(MethodProxy).get_default().proxy(signature.return_annotation)
        return getattr(self, '__proxy__', self.fn)

    @staticmethod
    def get_mpi_if_present(target: Any) -> "MPI":
        if hasattr(target, '__mpi__'):
            return target.__mpi__

        return MPI()


def mpi(fn: str = None, *, name='', version='', proto='grpc', codec='json', asyncable=True,
        timeout=10000, retries=3, node='', inst='', zone='', cluster='', cell='', group='', address='',
        encrypt=False) -> T:
    return MPI(fn, name, version, proto, codec, asyncable, timeout, retries, node, inst, zone, cluster, cell, group,
               address, encrypt)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Type, Any, Generic

from mesh.macro.ark import ark, T


class SPI(Generic[T]):

    def __init__(self, fn: T = None, name='', pattern='', priority=0, prototype=False):
        """
        Metadata annotation for Serial Peripheral Interface. Can be used with {@link ServiceLoader#load(Class)}
        or dependency injection at compile time and runtime time.
        """
        self.name = name
        self.pattern = pattern
        self.priority = priority
        self.prototype = prototype
        if name != '':
            self.name = name
            return
        if fn is None:
            return
        if type(fn) is str:
            self.name = fn
            return
        self.name = str(type(fn))
        self.kind = fn

    def __call__(self, *args, **kwargs) -> T:
        if not hasattr(self, 'kind'):
            self.kind = args[0]
        self.kind.__spi__ = self
        ark.register(spi, self.name, self.kind, self)
        return self.kind

    @staticmethod
    def get_macro(kind: Type[Any]) -> "SPI":
        if hasattr(kind, '__spi__'):
            return kind.__spi__
        return SPI()


def spi(fn: T = None, *, name='', pattern='', priority=0, prototype=False):
    return SPI(fn, name, pattern, priority, prototype)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Dict

from mesh.macro import index, serializable


@serializable
class Versions:

    @index(0)
    def version(self) -> str:
        """
        Platform version.

        main.sub.feature.bugfix like 1.5.0.0
        :return:
        """
        return ""

    @index(5)
    def infos(self) -> Dict[str, str]:
        """
        Network modules info.
        :return:
        """
        return {}

    def compare(self, v: str) -> int:
        """
        Is the version less than the appointed version.
        :param v: compared version
        :return: -1 true less than, 0 equals, 1 gather than
        """
        if not self.version or "" == self.version:
            return -1
        if not v or "" == v:
            return 1

        rvs = self.version.split(".")
        vs = v.split(".")
        for idx, value in enumerate(rvs):
            if vs.__len__() <= idx:
                return 0
            if vs[idx] == "*":
                continue
            if rvs[idx] > vs[idx]:
                return 1
            if rvs[idx] < vs[idx]:
                return -1
        return 0



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod

import mesh.log as log
from mesh.cause import MeshCode, Codeable
from mesh.codec import Codec
from mesh.context import Mesh
from mesh.kinds import Outbound, MeshFlag, Service
from mesh.macro import spi, ServiceLoader, Cause
from mesh.mpc.eden import Eden
from mesh.mpc.invoker import ServiceInvocation, Execution
from mesh.mpc.urn import URN
from mesh.prsim import Context

PROVIDER = "provider"
CONSUMER = "consumer"


@spi(PROVIDER)
class Transporter(ABC):

    @abstractmethod
    def transport(self, ctx: Context, urn: str, buff: bytes) -> bytes:
        """
        Transport the stream.
        :param urn:
        :param buff:
        :return:
        """
        pass

    @staticmethod
    def get_codec(flag: str) -> Codec:
        codec = ServiceLoader.load(Codec).get(MeshFlag.of_code(flag).name)
        if codec:
            return codec
        return ServiceLoader.load(Codec).get_default()


@spi(PROVIDER)
class ProviderTransporter(Transporter):

    def transport(self, ctx: Context, urn: str, buff: bytes) -> bytes:
        return Mesh.context_safe(lambda: self.transport_safe(ctx, urn, buff))

    def transport_safe(self, ctx: Context, uname: str, buff: bytes) -> bytes:
        try:
            Mesh.context().rewrite_context(ctx)
            Mesh.context().rewrite_urn(uname)
            urn = URN.parse(uname)
            eden = ServiceLoader.load(Eden).get_default()
            codec = self.get_codec(urn.flag.codec)
            execution = eden.infer(uname)
            if execution is None:
                outbound = Outbound()
                outbound.code = MeshCode.NO_SERVICE.code
                outbound.message = f"No mpi named {urn.name}."
                return codec.encode(outbound)
            return self.service(urn, codec, execution, buff)
        except BaseException as e:
            trace_id = Mesh.context().get_trace_id()
            span_id = Mesh.context().get_span_id()
            log.error(f"{trace_id}#{span_id} Invoke service {uname} with error.", e)
            outbound = Outbound()
            outbound.code = MeshCode.SYSTEM_ERROR.code
            outbound.message = str(e)
            outbound.cause = Cause.of(e)
            return ServiceLoader.load(Eden).get_default().encode(outbound)

    @staticmethod
    def service(urn: URN, codec: Codec, execution: Execution[Service], buff: bytes) -> bytes:
        returns = execution.reflect()
        try:
            parameters = codec.decode(buff, execution.intype())
            invocation = ServiceInvocation(execution, execution.inspect(), parameters, execution, urn)
            result = execution.run(invocation)
            returns.set_code(MeshCode.SUCCESS.get_code())
            returns.set_message(MeshCode.SUCCESS.get_message())
            returns.set_content(result)
            return codec.encode(returns)
        except BaseException as e:
            if isinstance(e, Codeable):
                returns.set_code(e.get_code())
                returns.set_message(e.get_message())
            else:
                returns.set_code(MeshCode.SYSTEM_ERROR.code)
                returns.set_message(str(e))
            returns.set_cause(Cause.of(e))
            return codec.encode(returns)


@spi(CONSUMER)
class ConsumerTransporter(Transporter):

    def transport(self, ctx: Context, urn: str, buff: bytes) -> bytes:
        Mesh.context().rewrite_context(ctx)
        return b'unknown'



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import json
from typing import Type

from mesh.codec.codec import Codec, T
from mesh.macro import spi, Compatible

Json = "json"


@spi(Json)
class JsonCodec(Codec):

    def encode(self, value: T) -> bytes:
        default = value.encode if value and hasattr(value, 'encode') else None
        return json.dumps(value, default=default).encode('UTF-8')

    def decode(self, value: bytes, kind: Type[T]) -> T:
        if value is None:
            return None

        vt = Compatible.get_args(kind)
        if vt.__len__() < 1:
            return json.loads(value, cls=kind)

        if issubclass(vt[0], (complex, int, float, bool, str, set, tuple, list, iter, bytes, dict)):
            return json.loads(value)

        return json.loads(value, cls=vt[0] if kind else None)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import index, serializable


@serializable
class Reference:

    @index(0)
    def urn(self) -> str:
        return ''

    @index(5)
    def namespace(self) -> str:
        return ''

    @index(10)
    def name(self) -> str:
        return ''

    @index(15)
    def version(self) -> str:
        return ''

    @index(20)
    def proto(self) -> str:
        return ''

    @index(25)
    def codec(self) -> str:
        return ''

    @index(30)
    def asyncable(self) -> bool:
        return True

    @index(35)
    def timeout(self) -> int:
        return 5000

    @index(40)
    def retries(self) -> int:
        return 5

    @index(45)
    def node(self) -> str:
        return ''

    @index(50)
    def inst(self) -> str:
        return ''

    @index(55)
    def zone(self) -> str:
        return ''

    @index(60)
    def cluster(self) -> str:
        return ''

    @index(65)
    def cell(self) -> str:
        return ''

    @index(70)
    def group(self) -> str:
        return ''

    @index(75)
    def address(self) -> str:
        return ''



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Type

from mesh.codec import Codec
from mesh.macro import T, spi

Xml = "xml"


@spi(Xml)
class XmlCodec(Codec):

    def encode(self, value: T) -> bytes:
        pass

    def decode(self, value: bytes, kind: Type[T]) -> T:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mesh',
 'mesh.asm',
 'mesh.boost',
 'mesh.cause',
 'mesh.codec',
 'mesh.context',
 'mesh.environ',
 'mesh.examples',
 'mesh.grpx',
 'mesh.http',
 'mesh.ioc',
 'mesh.kinds',
 'mesh.log',
 'mesh.macro',
 'mesh.metrics',
 'mesh.mpc',
 'mesh.prsim',
 'mesh.runtime',
 'mesh.schema',
 'mesh.system',
 'mesh.test',
 'mesh.test.boost',
 'mesh.test.codec',
 'mesh.test.grpx',
 'mesh.test.ioc',
 'mesh.test.macro',
 'mesh.test.metrics',
 'mesh.test.mpc',
 'mesh.test.prsim',
 'mesh.test.tool',
 'mesh.tool']

package_data = \
{'': ['*']}

install_requires = \
['grpcio>=1.43.0,<2.0.0', 'protobuf>=3.14.0,<4.0.0', 'psutil>=5.9.1,<6.0.0']

setup_kwargs = {
    'name': 'imesh',
    'version': '0.0.7',
    'description': 'A lightweight, distributed, relational network architecture for MPC.',
    'long_description': '# Mesh Python Client\n\n[![Build Status](https://travis-ci.org/ducesoft/babel.svg?branch=master)](https://travis-ci.org/ducesoft/babel)\n[![Financial Contributors on Open Collective](https://opencollective.com/babel/all/badge.svg?label=financial+contributors)](https://opencollective.com/babel) [![codecov](https://codecov.io/gh/babel/babel/branch/master/graph/badge.svg)](https://codecov.io/gh/babel/babel)\n![license](https://img.shields.io/github/license/ducesoft/babel.svg)\n\n中文版 [README](README_CN.md)\n\n## Introduction\n\nMesh is a standard implementation for [Private Transmission Protocol](Specifications.md) specification.\n\nMesh Python develop kits base on Python3.6. Recommend use [poetry](https://github.com/python-poetry/poetry) to manage\ndependencies.\n\n## Features\n\nAs an open source Internet of Data infrastructure develop kits, Mesh has the following core functions:\n\n* Minimal kernel with SPI plugin architecture, everything is replacement.\n* Support full stack of service mesh architecture.\n* Support full stack of service oriented architecture.\n* Support transport with TCP, HTTP, or other RPC protocols.\n* Support rich routing features.\n* Support reliable upstream management and load balancing capabilities.\n* Support network and protocol layer observability.\n* Support mTLS and protocols on TLS.\n* Support rich extension mechanism to provide highly customizable expansion capabilities.\n* Support process smooth upgrade.\n\n## Get Started\n\n```bash\npoetry add imesh\n```\n\nor\n\n```bash\npip install imesh\n```\n\n### RPC\n\nDeclared rpc interface Facade.\n\n```python\n\nfrom abc import ABC, abstractmethod\n\nfrom mesh import spi, mpi\n\n\n@spi("mesh")\nclass Tokenizer(ABC):\n\n    @abstractmethod\n    @mpi("mesh.trust.apply")\n    def apply(self, kind: str, duration: int) -> str:\n        """\n        Apply a node token.\n        :param kind:\n        :param duration:\n        :return:\n        """\n        pass\n\n    @abstractmethod\n    @mpi("mesh.trust.verify")\n    def verify(self, token: str) -> bool:\n        """\n        Verify some token verifiable.\n        :param token:\n        :return:\n        """\n        pass\n```\n\nDeclared rpc service Implement.\n\n```python\n\nfrom mesh import mps, Tokenizer\n\n\n@mps\nclass MeshTokenizer(Tokenizer):\n\n    def apply(self, kind: str, duration: int) -> str:\n        return "foo"\n\n    def verify(self, token: str) -> bool:\n        return True\n```\n\nRemote reference procedure call.\n\n```python\n\nfrom mesh import mpi, Tokenizer\n\n\nclass Component:\n\n    @mpi\n    def tokenizer(self) -> Tokenizer:\n        pass\n\n    def invoke(self) -> bool:\n        token = self.tokenizer().apply(\'PERMIT\', 1000 * 60 * 5)\n        return self.tokenizer().verify(token)\n\n\n```\n\n### Transport\n\nTransport is a full duplex communication stream implement.\n\n```python\nimport mesh\nfrom mesh import Mesh, log, ServiceLoader, Transport, Routable\nfrom mesh.prsim import Metadata\n\n\ndef main():\n    mesh.start()\n\n    transport = Routable.of(ServiceLoader.load(Transport).get("mesh"))\n    session = transport.with_address("10.99.1.33:570").local().open(\'session_id_008\', {\n        Metadata.MESH_VERSION.key(): \'\',\n        Metadata.MESH_TECH_PROVIDER_CODE.key(): \'LX\',\n        Metadata.MESH_TRACE_ID.key(): Mesh.context().get_trace_id(),\n        Metadata.MESH_TOKEN.key(): \'x\',\n        Metadata.MESH_SESSION_ID.key(): \'session_id_008\',\n        Metadata.MESH_TARGET_INST_ID.key(): \'JG0100000100000000\',\n    })\n    for index in range(100):\n        inbound = f"节点4发送给节点5报文{index}"\n        log.info(f"节点4发送:{inbound}")\n        session.push(inbound.encode(\'utf-8\'), {}, "topic")\n        outbound = session.pop(10000, "topic")\n        if outbound:\n            log.info(f"节点4接收:{outbound.decode(\'utf-8\')}")\n\n```\n',
    'author': 'coyzeng',
    'author_email': 'coyzeng@gmail.com',
    'maintainer': 'coyzeng',
    'maintainer_email': 'coyzeng@gmail.com',
    'url': 'https://mesh.github.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


from typing import Dict, Iterator, Any

import grpc

from mesh.grpx.marshaller import GrpcMarshaller
from mesh.macro import ServiceLoader
from mesh.mpc import Transporter, PROVIDER, MeshContext
from mesh.prsim import Context, Metadata


class GrpcBindableService(grpc.GenericRpcHandler):

    def __init__(self):
        self.marshaller = GrpcMarshaller()
        self.handlers = grpc.method_handlers_generic_handler("mesh-rpc", self.grpc_handlers())

    def grpc_handlers(self) -> Dict[str, grpc.RpcMethodHandler]:
        return {
            "v1": grpc.stream_stream_rpc_method_handler(
                self.stream_stream,
                request_deserializer=self.marshaller.deserialize,
                response_serializer=self.marshaller.serialize,
            ),
        }

    def service(self, handler_call_details):
        return self.handlers.service(handler_call_details)

    @staticmethod
    def stream_stream(iterator, ctx: grpc.ServicerContext) -> Iterator[Any]:
        return Transformer(iterator, ctx)


class Transformer(Iterator[Any]):

    def __int__(self, iterator: Iterator[Any], ctx: grpc.ServicerContext):
        self.iterator = iterator
        self.grpc_ctx = ctx
        self.transporter = ServiceLoader.load(Transporter).get(PROVIDER)
        self.mesh_ctx = self.context(ctx)

    def __next__(self) -> Any:
        buff = next(self.iterator)
        return self.transporter.transport(self.mesh_ctx, self.mesh_ctx.get_urn(), buff)

    def on_error(self, e: BaseException):
        """"""
        pass

    def on_complete(self):
        """"""
        pass

    @staticmethod
    def context(ctx: grpc.ServicerContext) -> Context:
        metadata = ctx.invocation_metadata()
        if metadata is None:
            return MeshContext.create()
        mtx = MeshContext()
        for (name, value) in metadata:
            std_name = name.replace("_", "-").lower()
            if std_name == Metadata.MESH_URN.key() or std_name == 'authority':
                mtx.urn = value
                mtx.attachments[Metadata.MESH_URN.key()] = value
                continue
            if std_name == Metadata.MESH_TRACE_ID.key():
                mtx.trace_id = value
                mtx.attachments[Metadata.MESH_TRACE_ID.key()] = value
                continue
            if std_name == Metadata.MESH_SPAN_ID.key():
                mtx.span_id = value
                mtx.attachments[Metadata.MESH_SPAN_ID.key()] = value
                continue
            mtx.attachments[std_name] = value
        return mtx



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from asyncio import Future
from typing import Any

import mesh.log as log
from mesh.context import Mesh
from mesh.macro import T, ServiceLoader
from mesh.mpc.eden import Eden
from mesh.mpc.filter import Filter, PROVIDER
from mesh.mpc.invoker import Invoker, Invocation


class ServiceInvoker(Invoker[T]):

    def __init__(self, service: Any):
        self.service = service

    def run(self, invocation: Invocation) -> Any:
        result = invocation.get_inspector().invoke(self.service, invocation.get_arguments())
        if not isinstance(result, Future):
            return result
        try:
            eden = ServiceLoader.load(Eden).get_default()
            execution = eden.infer(Mesh.context().get_urn())
            timeout = execution.schema().timeout
            return result.result(timeout=max(timeout, 3000))
        except BaseException as e:
            log.error(f'Invoke {Mesh.context().get_urn()} fault because of {e}')
            raise e


class ServiceInvokeHandler(Invoker):
    """

    """

    def __init__(self, service: Any):
        self.invoker = Filter.composite(ServiceInvoker(service), PROVIDER)

    def run(self, invocation: Invocation) -> Any:
        return self.invoker.run(invocation)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Dict

from mesh.macro import index, serializable


@serializable
class Service:
    """
    Service definition.
            service = Service()
            service.urn = ""
            service.namespace = ""
            service.name = ""
            service.version = version
            service.proto = proto
            service.codec = codec
            service.asyncable = asyncable
            service.timeout = 3000
            service.retries = 3
            service.node = ""
            service.inst = ""
            service.zone = ""
            service.cluster = ""
            service.cell = ""
            service.group = ""
            service.address = ""
            service.kind = ""
            service.lang = ""
            service.attrs = {}
    """

    @index(0)
    def urn(self) -> str:
        """
        Uniform domain name.
        """
        pass

    @index(5)
    def namespace(self) -> str:
        """
        Service namespace.
        """
        pass

    @index(10)
    def name(self) -> str:
        """
        Service name.
        """
        pass

    @index(15)
    def version(self) -> str:
        """
        Service version.
        """
        pass

    @index(20)
    def proto(self) -> str:
        """
        Service net protocol.
        """
        pass

    @index(25)
    def codec(self) -> str:
        """
        Service payload codec protocol.
        """
        pass

    @index(30)
    def asyncable(self) -> bool:
        """
        Service serve asyncable.
        """
        pass

    @index(35)
    def timeout(self) -> int:
        """
        Service serve timeout in mill seconds.
        """
        pass

    @index(40)
    def retries(self) -> int:
        """
        Service serve retry times.
        """
        pass

    @index(45)
    def node(self) -> str:
        """
        Service customized zone.
        """
        pass

    @index(50)
    def inst(self) -> str:
        """
        Service customized institution.
        """
        pass

    @index(55)
    def zone(self) -> str:
        """
        Service customized zone.
        """
        pass

    @index(60)
    def cluster(self) -> str:
        """
        Service customized cluster.
        """
        pass

    @index(65)
    def cell(self) -> str:
        """
        Service customized cell.
        """
        pass

    @index(70)
    def group(self) -> str:
        """
        Service customized group.
        """
        pass

    @index(75)
    def group(self) -> str:
        """
        Service customized sets.
        """
        pass

    @index(80)
    def address(self) -> str:
        """
        Service customized address.
        """
        pass

    @index(85)
    def kind(self) -> str:
        """
        Service kind.
        """
        pass

    @index(90)
    def lang(self) -> str:
        """
        Service lang.
        """
        pass

    @index(95)
    def attrs(self) -> Dict[str, str]:
        """
        Service customized attrs.
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod
from typing import Dict, List

from mesh.kinds import Versions
from mesh.macro import spi, mpi


@spi("mesh")
class Builtin(ABC):

    @abstractmethod
    @mpi("${mesh.name}.builtin.doc")
    def doc(self, name: str, formatter: str) -> str:
        """
        Export the documents.
        :param name:
        :param formatter:
        :return:
        """
        pass

    @abstractmethod
    @mpi("${mesh.name}.builtin.version")
    def version(self) -> Versions:
        """
        Get the builtin application version.
        :return:
        """
        pass

    @abstractmethod
    @mpi("${mesh.name}.builtin.debug")
    def debug(self, features: Dict[str, str]):
        """
        LogLevel set the application log level.
        :return:
        """
        pass

    @abstractmethod
    @mpi("${mesh.name}.builtin.stats")
    def stats(self, features: List[str]) -> Dict[str, str]:
        """
        Health check stats.
        :param features:
        :return:
        """
        pass
